from . import validation as validation
from . import graph as graphs
from . import graphviz as graphviz
from . import description as description
from . import identity as identity
from . import interface as interface

# class IdentityMethods:

#     # hash = hash_methods.hash
#     # hash_elems = hash_methods.hash_elems
#     _hash = hash_methods._hash
#     _hash_elems = hash_methods._hash_elems
#     find_isomorphism = hash_methods.find_isomorphism
#     canonical = hash_methods.canonical
#     reindex = hash_methods.reindex

# class TaxonomyMethods:

#     # assert_poset = taxonomy_methods.assert_poset
#     # is_poset = taxonomy_methods.is_poset
#     # assert_lattice = taxonomy_methods.assert_lattice
#     # is_lattice = taxonomy_methods.is_lattice
#     # is_distributive = taxonomy_methods.is_distributive
#     # explain_non_distributive = taxonomy_methods.explain_non_distributive
#     # assert_distributive = taxonomy_methods.assert_distributive
#     # is_modular = taxonomy_methods.is_modular
#     # explain_non_modular = taxonomy_methods.explain_non_modular
#     # assert_modular = taxonomy_methods.assert_modular
#     pass

# class GraphMethods:

#     # bottom = graph_methods.bottom
#     # top = graph_methods.top
#     # heights = graph_methods.heights
#     # height = graph_methods.height
#     subgraph = graph_methods.subgraph
#     toposort = graph_methods.toposort
#     toporank = graph_methods.toporank
#     independent_components = graph_methods.independent_components
#     bottoms = graph_methods.bottoms
#     tops = graph_methods.tops
#     non_bottoms = graph_methods.non_bottoms
#     non_tops = graph_methods.non_tops

# class MetaMethods:
#     # meta_irreducibles = meta_methods.meta_irreducibles
#     # meta_downsets = meta_methods.meta_downsets
#     # meta_endomorphisms = meta_methods.meta_endomorphisms
#     # meta_JE = meta_methods.meta_JE
#     # meta_JJ = meta_methods.meta_JJ
#     # upside_down = meta_methods.upside_down
#     # power = meta_methods.power
#     pass

# class TestingMethods:
#     # _test_iters_diff = testing_methods._test_iters_diff
#     # _test_iters = testing_methods._test_iters
#     # _test_counts = testing_methods._test_counts
#     # _test_summary = testing_methods._test_summary
#     # _test_assert_distributive = testing_methods._test_assert_distributive
#     # test_iter_f_monotone = testing_methods.test_iter_f_monotone
#     # test_iter_f_monotone_bottom = testing_methods.test_iter_f_monotone_bottom
#     # test_iter_f_lub = testing_methods.test_iter_f_lub
#     # test_iter_f_lub_pairs = testing_methods.test_iter_f_lub_pairs
#     # test_iter_f_lub_distributive = testing_methods.test_iter_f_lub_distributive
#     # test_count_f_lub_distributive = testing_methods.test_count_f_lub_distributive
#     pass

# class BinaryOperatorMethods:
#     # add_poset = binary_operator_methods.add_poset
#     # mul_poset = binary_operator_methods.mul_poset
#     # or_poset = binary_operator_methods.or_poset
#     # and_poset = binary_operator_methods.and_poset
#     # add_number = binary_operator_methods.add_number
#     # mul_number = binary_operator_methods.mul_number
#     # or_number = binary_operator_methods.or_number
#     # and_number = binary_operator_methods.and_number
#     # _operation_number = binary_operator_methods._operation_number
#     pass

# class EndomorphismMethods:
#     iter_f_all = endomorphisms.iter_f_all
#     num_f_all = endomorphisms.num_f_all
#     iter_f_all_bottom = endomorphisms.iter_f_all_bottom
#     num_f_all_bottom = endomorphisms.num_f_all_bottom
#     f_is_monotone = endomorphisms.f_is_monotone
#     f_meet = endomorphisms.f_meet
#     _as_external_lattice = endomorphisms._as_external_lattice
#     iter_f_monotone_bruteforce = endomorphisms.iter_f_monotone_bruteforce
#     iter_f_monotone_bottom_bruteforce = endomorphisms.iter_f_monotone_bottom_bruteforce
#     iter_f_monotone = endomorphisms.iter_f_monotone
#     iter_f_lub_bruteforce = endomorphisms.iter_f_lub_bruteforce
#     iter_f_monotone_restricted = endomorphisms.iter_f_monotone_restricted
#     _iter_f_monotone_restricted = endomorphisms._iter_f_monotone_restricted
#     _toposort_children = endomorphisms._toposort_children
#     iter_f_monotone_bottom = endomorphisms.iter_f_monotone_bottom
#     irreducible_components = endomorphisms.irreducible_components
#     _interpolate_funcs = endomorphisms._interpolate_funcs
#     iter_f_irreducibles_monotone_bottom = endomorphisms.iter_f_irreducibles_monotone_bottom
#     iter_f_irreducibles_monotone = endomorphisms.iter_f_irreducibles_monotone
#     f_is_lub = endomorphisms.f_is_lub
#     f_is_lub_pairs = endomorphisms.f_is_lub_pairs
#     iter_f_lub_pairs_bruteforce = endomorphisms.iter_f_lub_pairs_bruteforce
#     iter_f_lub_pairs = endomorphisms.iter_f_lub_pairs
#     iter_f_lub = endomorphisms.iter_f_lub
#     num_f_lub_pairs = endomorphisms.num_f_lub_pairs
#     count_f_lub_pairs_bruteforce = endomorphisms.count_f_lub_pairs_bruteforce
#     num_f_lub = endomorphisms.num_f_lub
#     count_f_lub = endomorphisms.count_f_lub
#     count_f_lub_bruteforce = endomorphisms.count_f_lub_bruteforce
#     f_is_lub_of_irreducibles = endomorphisms.f_is_lub_of_irreducibles

#     def iter_f_lub_distributive(self):
#         pass

#     def count_f_lub_distributive(self):
#         pass

#         def num(i: int):
#             pass

# class GenerationMethods:
#     all_latices = generation.all_latices
#     iter_all_latices = generation.iter_all_latices
#     random_poset = generation.random_poset
#     random_lattice_czech = generation.random_lattice_czech

# class InitializationMethods:

#     # total = initialization.total
#     # powerset = initialization.powerset
#     # Mn = initialization.Mn
#     # from_parents = initialization.from_parents
#     # from_children = initialization.from_children
#     # from_down_edges = initialization.from_down_edges
#     # from_up_edges = initialization.from_up_edges
#     # from_lambda = initialization.from_lambda
#     # child_to_dist = initialization.child_to_dist



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, List, Optional
if TYPE_CHECKING:
    from ..base import Lattice, Poset, Relation
from typing import Sequence, Tuple


def graphviz(
    n: int,
    edges: Sequence[Tuple[int, int]],
    labels: Sequence[str],
    extra_edges: Optional[Dict[str, List[Tuple[int, int]]]] = None,
    save: Optional[str] = None,
):
    'Show graph using graphviz. blue edges are extra edges'
    from pydotplus import graph_from_edges
    from pydotplus.graphviz import Node, Edge

    black_color = '#aaaaaa'

    g = graph_from_edges([], directed=True)
    g.set_rankdir('TB')  # type:ignore
    for i in range(n):
        style = {
            'margin': 0,
            'width': 0.35,
            'shape': 'circle',
        }
        g.add_node(Node(i, label=f'"{labels[i]}"', **style))

    for i, j in edges:
        style = {
            'dir': 'none',
            'color': black_color,
            'arrowsize': 0.5,
        }
        g.add_edge(Edge(i, j, **style))

    if extra_edges is not None:
        for color, color_edges in extra_edges.items():
            for i, j in color_edges:
                style = {
                    'color': color,
                    'constraint': 'false',
                    'arrowsize': 0.65,
                }
                g.add_edge(Edge(i, j, **style))

    png = g.create_png()  # type:ignore

    if save is None:
        import builtins
        if hasattr(builtins, '__IPYTHON__'):
            from IPython.display import display
            from IPython.display import Image
            img = Image(png)
            display(img)
        else:
            from io import BytesIO
            from PIL import Image
            img = Image.open(BytesIO(png))
            img.show()
    else:
        with open(save, 'wb') as f:
            f.write(png)
    return


def show(
    P: Poset,
    f=None,
    method='auto',
    labels=None,
    save=None,
):
    '''
    Use graphviz to display or save P as a Hasse diagram.
    The argument "method" (string) only affects visualization
    of the endomorphism f (if given). It can be
        - arrows: blue arrow from each node i to f[i]
        - labels: replace the label i of each node with f[i]
        - labels_bottom: (no label at i if f[i]=bottom)
        - arrows_bottom: (no arrow at i if f[i]=bottom)
        - auto: 'arrows_bottom' if P is a lattice and f preserves lub. 'arrows' otherwise.
    Hidding bottom is only allowed if P.bottom makes sense.
    '''
    methods = ('auto', 'labels', 'arrows', 'labels_bottom', 'arrows_bottom')
    assert method in methods, f'Unknown method "{method}"'

    L = P.as_lattice(check=False)
    if method == 'auto' and f is not None:
        method = 'arrows'
    n = P.n
    child = P.child
    extra: List[Tuple[int, int]] = []
    if labels is None:
        labels = P._labels
    if f is not None:
        enabled = not method.endswith('_bottom')
        ok = lambda fi: enabled or fi != L.bottom
        if method.startswith('arrows'):
            extra = [(i, int(f[i])) for i in range(n) if ok(f[i])]
        else:
            gr = [[] for _ in range(n)]
            for i in range(n):
                if ok(f[i]):
                    gr[f[i]].append(i)
            labels = [','.join(map(str, l)) for l in gr]

    if extra and L.is_lattice and f is not None and L.f_preserves_lub(f):
        extra_edges = {'darkgreen': extra}
    else:
        extra_edges = {'blue': extra}

    edges = [(i, j) for i in range(n) for j in range(n) if child[j, i]]
    graphviz(n, edges, labels=labels, extra_edges=extra_edges, save=save)
    return



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, List

if TYPE_CHECKING:
    from ..base import Lattice, Poset, Relation, DistributiveLattice, ModularLattice

import numpy as np


class ValidationError(AssertionError):

    _message = 'Validation failed'

    @classmethod
    def assert_false(cls, why_false: str):
        if why_false:
            raise cls(why_false)

    @classmethod
    def capture(cls, method: Callable):
        try:
            method()
        except ValidationError:
            return False
        return True


# Not poset
class MatrixFormatError(ValidationError):
    _message = 'Given matrix relation has incorrect format'


def validate_matrix_format(R: Relation):
    rel = R.leq
    try:
        shape = tuple(rel.shape)
        assert len(shape) == 2, f'Shape {shape} found'
        n = shape[0]
        assert shape == (n, n), f'Shape {shape} found'
        assert rel.dtype == bool, f'Dtype {rel.dtype} found'
        assert rel.flags.writeable == False, 'flags.writeable is True'
    except AssertionError as e:
        message = ''.join([
            'The relation matrix must be a 2-dimensional boolean numpy array',
            ', squared and read-only (i.e. rel.flags.writeable = False).',
            f'\n{e}'
        ])
        raise MatrixFormatError(message)
    R.n = rel.shape[0]
    R.leq = rel
    return


class NotPosetError(ValidationError):
    _message = 'Given relation is not a poset'


class NotTransitive(NotPosetError):
    _message = 'Given relation is not transitive'


class NotReflexive(NotPosetError):
    _message = 'Given relation is not reflexive (i<=i)'


class NotAntisymmetric(NotPosetError):
    _message = ('Given relation is not antisymmetric '
                '(antisymmetry: i<=j and j<=i imply j=i)\n'
                'There is a cycle')


def assert_is_reflexive(R: Relation):
    rel = R.leq
    I, = np.where(~rel[np.diag_indices_from(rel)])
    why = I.size and f'Not reflexive: rel[{I[0]},{I[0]}] is False'
    NotReflexive.assert_false(why)


def assert_is_antisymmetric(R: Relation):
    rel = R.leq
    eye = np.identity(R.n, dtype=np.bool_)
    I, J = np.where(rel & rel.T & ~eye)
    why = I.size and f'Not antisymmetric: cycle {I[0]}<={I[1]}<={I[0]}'
    NotAntisymmetric.assert_false(why)


def assert_is_transitive(R: Relation):
    rel = R.leq
    rel2 = np.matmul(rel, rel)
    I, J = np.where(((~rel) & rel2))
    why = I.size and (
        f'Not transitive: rel[{I[0]},{J[0]}] is False but there is a path')
    NotTransitive.assert_false(why)


def is_reflexive(R: Relation):
    method = lambda: assert_is_reflexive(R)
    return ValidationError.capture(method)


def is_antisymmetric(R: Relation):
    method = lambda: assert_is_antisymmetric(R)
    return ValidationError.capture(method)


def is_transitive(R: Relation):
    method = lambda: assert_is_transitive(R)
    return ValidationError.capture(method)


def assert_is_poset(P: Poset):
    assert_is_reflexive(P)
    assert_is_antisymmetric(P)
    assert_is_transitive(P)
    return


class NoBottoms(NotTransitive):
    pass


class NoTops(NotTransitive):
    pass


# Not lattice
class NotLattice(ValidationError):
    _message = 'Given poset is not a lattice'


class LUB_Inconsistency(ValidationError):

    def __init__(self, i: int, j: int):
        self.args = (i, j)


def assert_is_lattice(L: Lattice):
    if L.n == 0:
        return
    try:  # check if duck says quack
        L.bottom
        L.top
        L.lub
        return
    except LUB_Inconsistency as e:
        i, j = e.args
    # Find the error and explain it to the user
    n = L.n
    leq = L.leq
    above = [k for k in range(n) if leq[i, k] and leq[j, k]]
    below = [k for k in range(n) if leq[k, i] and leq[k, j]]
    try:
        assert above, f'Not a lattice: {i} lub {j} => (no common ancestor)'
        assert below, f'Not a lattice: {i} glb {j} => (no common descendant)'
        lub = min(above, key=lambda k: sum(leq[:, k]))
        glb = max(below, key=lambda k: sum(leq[:, k]))
        for x in above:
            assert leq[lub, x], f'Not a lattice: {i} lub {j} => {lub} or {x}'
        for x in below:
            assert leq[x, glb], f'Not a lattice: {i} glb {j} => {glb} or {x}'
        assert False, 'Unkown reason'
    except AssertionError as e:
        raise NotLattice(e)


class NotComplete(NotLattice):  # Because all finite lattices are complete
    _message = 'Given poset is not complete'


class NotUniqueBottom(NotComplete):
    _message = 'Given poset has multiple bottom elements'


class NotUniqueTop(NotComplete):
    _message = 'Given poset has multiple top elements'


def expect_unique_bottom(bottoms: List[int]):
    if not bottoms:
        raise NoBottoms()
    if len(bottoms) > 1:
        hook = lambda: f'Multiple bottoms found: {bottoms}'
        raise NotUniqueBottom(hook)
    return bottoms[0]


def expect_unique_top(tops: List[int]):
    if not tops:
        raise NoTops()
    if len(tops) > 1:
        hook = lambda: f'Multiple tops found: {tops}'
        raise NotUniqueTop(hook)
    return tops[0]


class NotDistributive(ValidationError):
    _message = 'Given lattice is not distributive'


def assert_is_distributive(self: DistributiveLattice):
    'Find i, j, k that violate distributivity. None otherwise'
    n = self.n
    lub = self.lub
    glb = self.glb
    for i in range(n):
        diff = glb[i, lub] != lub[np.ix_(glb[i, :], glb[i, :])]
        if diff.any():
            j, k = next(zip(*np.where(diff)))  # type:ignore
            raise NotDistributive(
                f'Non distributive lattice:\n'
                f'{i} glb ({j} lub {k}) = {i} glb {lub[j,k]} = '
                f'{glb[i,lub[j,k]]} != {lub[glb[i,j],glb[i,k]]} = '
                f'{glb[i,j]} lub {glb[i,k]} = ({i} glb {j}) lub ({i} glb {k})')
    return


class NotModular(ValidationError):
    _message = 'Given lattice is not distributive'


def assert_is_modular(self: ModularLattice):
    'Find i, j, k that violate modularity. None otherwise'
    n = self.n
    leq = self.leq
    lub = self.lub
    glb = self.glb
    for i in range(n):
        for j in range(n):
            if leq[i, j]:
                for k in range(n):
                    if glb[i, lub[k, j]] != lub[glb[i, k], j]:
                        raise NotModular(
                            f'Non modular lattice:\n'
                            f'{i} glb ({k} lub {j}) = {i} glb {lub[k,j]} = '
                            f'{glb[i,lub[k,j]]} != {lub[glb[i,k],j]} = '
                            f'{glb[i,k]} lub {j} = ({i} glb {k}) lub {j})')
    return


class RelationExceptions:
    MatrixFormatError = MatrixFormatError
    NotPosetError = NotPosetError
    NotTransitive = NotTransitive
    NotReflexive = NotReflexive
    NotAntisymmetric = NotAntisymmetric


class PosetExceptions:
    NotPosetError = NotPosetError
    NotTransitive = NotTransitive
    NotReflexive = NotReflexive
    NotAntisymmetric = NotAntisymmetric
    NotLattice = NotLattice
    LUB_Inconsistency = LUB_Inconsistency
    NotUniqueBottom = NotUniqueBottom
    NotUniqueTop = NotUniqueTop
    NotComplete = NotComplete
    NotDistributive = NotDistributive
    NoBottoms = NoBottoms
    NoTops = NoTops
    NotModular = NotModular



from setuptools import setup, find_packages
from avispa_lattices import VERSION

setup(
    name='avispa-lattices',
    version=VERSION,
    description='Toolbox for finite lattices',
    url='https://github.com/caph1993/avispa-lattices',
    author='Carlos Pinzón',
    author_email='caph1993@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'typing-extensions',
        'pillow',
        'ipython',
        'pyhash',
        'pydotplus',
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)


import itertools
from typing import Generator, Iterable, List, cast, TypeVar

F = TypeVar('F')


def product_list(*iterables: Iterable[F], repeat=1,
                 out=None) -> Generator[List[F], None, None]:
    'same as itertools.product, but mutates the output instead of making tuples'
    dims = [list(it) for it in iterables] * repeat
    n = len(dims)
    if out is not None:
        assert len(out) == n, f'Incompatible output shape'

    out = [None] * n if out is None else out
    out = cast(List[F], out)

    def backtrack(i):
        if i == n:
            yield out
        else:
            for x in dims[i]:
                out[i] = x
                yield from backtrack(i + 1)

    yield from backtrack(0)


def cartesian(*sizes: int):
    return product_list(*(range(size) for size in sizes))


from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, List, Tuple, Type, TypeVar, Union, cast
import numpy as np

from ..utils.numpy_types import npBoolMatrix, npUInt64Matrix

if TYPE_CHECKING:
    from ..base import Poset, Relation, Lattice
    CLS = TypeVar('CLS', type[Poset], type[Relation], type[Lattice])


def from_parents(cls: CLS, parents, labels=None, check=True):
    'create Poset from list: parents[i] = list of parents of i'
    children = cls.MD.graphs.parents_to_children(parents)
    return cls.from_children(children, labels, check)


def from_children(cls: CLS, children: List[List[int]], labels=None, check=True):
    'create Poset from list: children[i] = list of covers of i'
    n = len(children)
    child = np.zeros((n, n), dtype=bool)
    for pa in range(n):
        for ch in children[pa]:
            child[ch, pa] = True
    child.flags.writeable = False
    dist = cls.MD.graphs.child_to_dist(child, assume_poset=True)
    dist.flags.writeable = False
    leq = dist < n
    leq.flags.writeable = False
    return cls(leq, check, labels=labels, child=child, dist=dist)


def from_down_edges(cls: CLS, n, edges: Iterable[Tuple[int, int]], labels=None,
                    check=True):
    'create Poset of size n respecting all given relations (ancestor, descendant)'
    return cls.from_up_edges(n, [(j, i) for i, j in edges], labels, check)


def from_up_edges(cls: CLS, n, edges: Iterable[Tuple[int, int]], labels=None,
                  check=True):
    'create Poset of size n respecting all given relations (descendant, ancestor)'
    leq = np.zeros((n, n), dtype=bool)
    leq[np.diag_indices_from(leq)] = True
    for des, anc in edges:
        leq[des, anc] = True
    leq.flags.writeable = False
    R = cls(leq, check=False, labels=labels)
    R = R.transitive_closure()  # type: ignore
    #R = R.transitive_closure()
    P = R.copy(cls=cls, check=check)
    return P


def from_lambda(cls: CLS, elems, f_leq, labels=None, check=True):
    'create Poset with: leq[i,j] = f_leq(elems[i], elems[j])'
    m = len(elems)
    leq = np.zeros((m, m), dtype=bool)
    for i in range(m):
        for j in range(m):
            leq[i, j] = f_leq(elems[i], elems[j])
    leq.flags.writeable = False
    P = cls(leq, check=check, labels=labels)
    return P


from typing import Optional
import numpy as np


def random_state(seed: Optional[int] = None):
    if seed is None:
        seed = np.random.randint(0, 2**32)
    return np.random.RandomState(seed=seed)



from __future__ import annotations
import functools
from typing import Any, Callable, Generic, Iterable, List, Optional, Sequence, Set, Tuple, Type, TypeVar, Union, cast, overload
import numpy as np
import itertools
from .utils.methodtools import cached_property, cached_method
from .utils.algorithm_tarjan import Tarjan
from .utils.numpy_types import npBoolMatrix, npUInt64Matrix
from .utils.algorithm_floyd_warshall import floyd_warshall
from .utils.methodtools import implemented_at
from . import base_methods as MD
from .base_methods.validation import PosetExceptions

Endomorphism = List[int]
PartialEndomorphism = Union[List[Optional[int]], List[None], Endomorphism]
_T = TypeVar('_T')


class Relation:
    '''
    Class for boolean relation matrices intended mostly for asserting that
    a matrix relation can be used with the fully featured parent classes:
    Poset and Lattice.

    The attribute name leq (less or equal) is missleading for arbitrary relations
    on purpose, because it is mostly used in parent classes.
    '''

    n: int
    leq: npBoolMatrix
    labels: Optional[Sequence[str]]

    MD = MD

    # validation = validation
    # graph = graphs
    # graphviz = graphviz
    # description = description

    def __init__(self, leq: npBoolMatrix, check: bool,
                 labels: Optional[Sequence[str]] = None, **kwargs):
        self.n = len(leq)
        self.leq = leq
        self.labels = labels
        MD.validation.validate_matrix_format(self)
        for k, v in kwargs.items():
            self.__dict__[k] = v
        if check:
            self.validate()
        return

    def validate(self):
        pass  # All binary matrices are valid relations

    @cached_property
    def is_poset(self):
        method = lambda: self.as_poset(check=True)
        return MD.validation.ValidationError.capture(method)

    def as_poset(self, check: bool):
        return self.copy_as(Poset, check=check)

    def copy(self, check=False):
        return self.copy_as(self.__class__, check=check)

    def copy_as(self, cls: Type[_T], check=False) -> _T:
        kwargs = self.__dict__.copy()
        kwargs.pop('leq')
        kwargs.pop('labels', None)
        cpy = cls(self.leq, labels=self.labels, check=check, **kwargs)
        return cpy

    def copy_if(self, cls: Type[_T]) -> Optional[_T]:
        try:
            return self.copy_as(cls, check=True)
        except MD.validation.ValidationError:
            return None

    def reindex(self, rank: List[int], inverse=False):
        return MD.identity.reindex(self, rank, inverse=inverse)

    @property
    def _labels(self):
        return self.labels or [f'{i}' for i in range(self.n)]

    def relabel(self, labels: Optional[Sequence[str]] = None):
        'copy of self with different labels'
        if labels == self.labels:
            return self
        if labels is not None:
            n = self.n
            m = len(labels)
            assert m == n, f'{m} labels found. Expected {n}'
            non = [l for l in labels if not isinstance(l, str)]
            assert not non, f'non-string label found: {non[0]}'
        Q = self.copy()
        Q.labels = labels
        return Q

    @classmethod
    def total(cls, n: int):
        'total order of n elements'
        G = [[i - 1] if i > 0 else [] for i in range(n)]
        return cls.from_children(G, check=False)

    '''
    @section
        Interface methods
    '''

    @classmethod
    @implemented_at(MD.interface.from_parents)
    def from_parents(cls, *args, **kwargs):
        ...

    @classmethod
    @implemented_at(MD.interface.from_children)
    def from_children(cls, *args, **kwargs):
        ...

    @classmethod
    @implemented_at(MD.interface.from_down_edges)
    def from_down_edges(cls, *args, **kwargse):
        ...

    @classmethod
    @implemented_at(MD.interface.from_up_edges)
    def from_up_edges(cls, *args, **kwargse):
        ...

    @classmethod
    @implemented_at(MD.interface.from_lambda)
    def from_lambda(cls, *args, **kwargs):
        ...

    # def describe(self):
    #     self.show()
    #     print('Relation matrix:')
    #     print(self.leq.astype(int))
    #     print('Reflexive?', self.is_reflexive)
    #     print('Antisymmetric?', self.is_antisymmetric)
    #     print('Transitive?', self.is_transitive)
    #     return

    # def show(self, labels=None, save=None):
    #     'Display the relation using graphviz. Groups SCCs together'
    #     scc_components, scc_edges = self.scc_reduction()
    #     if labels is None:
    #         labels = [f'{i}' for i in range(self.n)]
    #     n = len(scc_components)
    #     labels = ['-'.join(labels[i] for i in I) for I in scc_components]
    #     return graphviz(n, edges=scc_edges, labels=labels, save=save)
    '''
    @section
        Graph operations
    '''

    def scc_reduction(self):
        n = self.n
        rel = self.leq
        G = [[] for _ in range(n)]
        for i in range(n):
            for j in range(n):
                if rel[i, j] and i != j:
                    G[i].append(j)
        return Tarjan(G).tarjan()

    def transitive_closure(self):
        dist = floyd_warshall(self.leq, infinity=self.n)
        rel = (dist < len(dist))
        rel.flags.writeable = False
        return self.__class__(rel, check=False)

    def __lt_pairs(self):
        for i in range(self.n):
            for j in range(self.n):
                if self.leq[i, j] and i != j:
                    yield i, j

    def f_is_monotone(self, f: Endomorphism):
        for i, j in self.__lt_pairs():
            if not self.leq[f[i], f[j]]:
                return False
        return True


class Poset(Relation):
    '''
    Hashable object that represents an inmutable finite partial order.
    Uses a matrix and hashing is invariant under permutations.

    Run print(Poset.usage) for details and usage examples.

    The main attributes (always present) are:
        n: size of the poset.
            The elements of the poset are range(n)
        leq: read only less-or-equal boolean nxn matrix:
            leq[i,j] is True if and only "i <= j" in the poset order
        
        labels: optional sequence of n strings. Only used for displaying
        _idx: optional permutation of 1..n (not the identity).
    '''

    def validate(self):
        super().validate()
        MD.validation.assert_is_poset(self)

    @cached_property
    def is_lattice(self):
        return self.copy_if(Lattice) is not None

    '''
    @section
        Methods for identity
    '''

    def __hash__(self):
        return self.hash

    @cached_property
    def hash(self):
        'hash number for this poset'
        return MD.identity.hasher(sorted(self.hash_elems))

    @cached_property
    def hash_elems(self):
        'hash for each element of the poset w.r.t. the poset order'
        return MD.identity._hash_elems(self, rounds=2, salt=0)

    def __eq__(self, other: Poset):
        'Equality up to isomorphism, i.e. up to reindexing'
        return MD.identity.find_isomorphism(self, other) is not None

    @cached_property
    def hash(self):
        'hash number for this poset'
        return MD.identity.hasher(sorted(self.hash_elems))

    @cached_property
    def canonical(self):
        rank = MD.identity.canonical_rank(self)
        P = self.reindex(rank)
        P.labels = None
        return P

    # __graph = methods_directory.GraphMethods
    # __testing = methods_directory.TestingMethods
    # __generation = methods_directory.GenerationMethods
    # __endomorphism = methods_directory.EndomorphismMethods
    # __initialization = methods_directory.InitializationMethods
    # __binary_operator = methods_directory.BinaryOperatorMethods
    '''
    @section
        Fundamental properties
    '''

    @cached_property
    def child(self) -> npBoolMatrix:
        '''
        nxn boolean matrix: transitive reduction of the poset.
        child[i,j] == True iff j covers i (with no elements inbetween)
        '''
        child = MD.graphs.transitive_reduction(self.leq)
        child.flags.writeable = False
        return child

    @cached_property
    def children(self) -> List[List[int]]:
        ''' top-down adjoint list (j in G[i] iff i covers j)'''
        n = self.n
        child = self.child
        return [[j for j in range(n) if child[j, i]] for i in range(n)]

    @cached_property
    def parents(self) -> List[List[int]]:
        '''bottom-up adjoint list (j in G[i] iff j covers i)'''
        n = self.n
        child = self.child
        return [[j for j in range(n) if child[i, j]] for i in range(n)]

    @cached_property
    def dist(self) -> npUInt64Matrix:
        '''
        Matrix of shortest distance from i upwards to j through parents
        n represents infinity.
        '''
        return MD.graphs.child_to_dist(self.child, assume_poset=True)

    '''
    @section
        Display methods
    '''

    @implemented_at(MD.graphviz.show)
    def show(self):
        ...

    def __repr__(self):
        return self.name

    @cached_property
    def name(self):
        return MD.description.name(self)

    @cached_property
    def upside_down(self):
        geq = self.leq.T
        Q = self.__class__(geq, check=False, labels=self.labels,
                           upside_down=self)
        return Q

    @cached_property
    def toposort_bottom_up(self):
        return MD.graphs.toposort_bottom_up(self)

    @cached_property
    def bottoms(self):
        return MD.graphs.bottoms(self)

    @cached_property
    def tops(self):
        return MD.graphs.tops(self)


class Lattice(Poset):

    def validate(self):
        super().validate()
        MD.validation.assert_is_lattice(self)

    @cached_property
    def is_distributive(self):
        return self.copy_if(DistributiveLattice) is not None

    @cached_property
    def is_modular(self):
        return self.copy_if(ModularLattice) is not None

    @cached_property
    def lub(self):
        'matrix of i lub j, i.e. i join j'
        n = self.n
        leq = self.leq
        lub_id = {tuple(leq[i, :]): i for i in range(n)}
        lub = np.zeros((n, n), int)
        for i in range(n):
            for j in range(n):
                above = tuple(leq[i, :] & leq[j, :])
                if above not in lub_id:
                    raise PosetExceptions.LUB_Inconsistency(i, j)
                lub[i, j] = lub_id[above]
        lub.flags.writeable = False
        return lub

    @property
    def glb(self):
        return self.upside_down.lub

    @cached_property
    def bottom(self):
        'unique bottom element of the Poset. Throws if not present'
        bottoms = MD.graphs.bottoms(self)
        return MD.validation.expect_unique_bottom(bottoms)

    @cached_property
    def top(self):
        'unique top element of the Poset. Throws if not present'
        tops = MD.graphs.tops(self)
        return MD.validation.expect_unique_top(tops)

    @cached_property
    def irreducibles(self):
        n = self.n
        children = self.children
        return tuple([i for i in range(n) if len(children[i]) == 1])

    @cached_property
    def irreducible_descendants(self):
        '''Irreducibles below x for each x'''
        Rn = range(self.n)
        I = self.irreducibles
        leq = self.leq
        return tuple([tuple([i for i in I if leq[i, x]]) for x in Rn])

    @cached_property
    def irreducible_components(self: Lattice):
        '''
        components of join irreducibles in toposort order and children
        lists for each component
        '''
        n = self.n
        if n <= 1:  # no join irreducibles at all
            return (0, [], [])
        graphs = self.MD.graphs
        irr = self.irreducibles
        sub = graphs.subgraph(self, irr)
        subcomps = graphs.independent_components(sub)
        m = len(subcomps)
        irrcomps = [[irr[j] for j in subcomps[i]] for i in range(m)]
        m_topo, m_children = zip(*(
            MD.graphs._toposort_children(self, irrcomps[i]) for i in range(m)))
        m_topo = cast(Tuple[List[int]], m_topo)
        m_children = cast(Tuple[List[List[int]]], m_children)
        return m, m_topo, m_children

    def lub_of_many(self, elems: Iterable[int]) -> int:
        return functools.reduce(
            function=lambda i, j: self.lub[i, j],
            sequence=elems,
            initial=self.bottom,
        )

    def glb_of_many(self, elems: Iterable[int]) -> int:
        return functools.reduce(
            function=lambda i, j: self.glb[i, j],
            sequence=elems,
            initial=self.top,
        )


class DistributiveLattice(Lattice):

    def validate(self):
        super().validate()
        MD.validation.assert_is_distributive(self)


class ModularLattice(Lattice):

    def validate(self):
        super().validate()
        MD.validation.assert_is_modular(self)

    # def f_preserves_lub(self, f: Endomorphism):
    #     return all(f[self.lub[i, j]] == self.lub[f[i], f[j]]
    #                for i in range(self.n)
    #                for j in range(self.n))

    # def f_lub(self, f: Endomorphism, g: Endomorphism) -> Endomorphism:
    #     return [self.lub[f[i], g[i]] for i in range(self.n)]

    # def f_glb_pointwise(self, f: Endomorphism, g: Endomorphism) -> Endomorphism:
    #     return [self.glb[f[i], g[i]] for i in range(self.n)]

    # def f_glb(self, f: Endomorphism, g: Endomorphism) -> Endomorphism:
    #     return self.f_preserving_lub(self.f_glb_pointwise(f, g))

    # def f_preserving_lub(self, f: Endomorphism,
    #                      budget: Optional[int] = None) -> Endomorphism:
    #     '''
    #     The famous algorithm that fixes iteratively the pairs of
    #     elements in the lattice that fail to satisfy the LUB axiom
    #     '''
    #     lub = self.lub
    #     leq = self.leq
    #     it = itertools.count() if budget is None else range(budget)
    #     for _ in it:
    #         f_prev = f.copy()
    #         for i in range(self.n):
    #             for j in range(self.n):
    #                 k = lub[i, j]
    #                 fi_lub_fj = lub[f[i], f[j]]
    #                 if fi_lub_fj == f[k]:
    #                     pass
    #                 elif leq[f[k], fi_lub_fj]:
    #                     f[k] = fi_lub_fj
    #                 else:
    #                     f[i] = lub[f[i], f[k]]
    #                     f[j] = lub[f[j], f[k]]
    #         if f == f_prev:
    #             break
    #     return f
